#include "libscio.hpp"
#include <unistd.h>
#include <cstring>

sc::io::memstreambuf::memstreambuf(char* mem, size_t size) {
	init(mem, size);
}

void sc::io::memstreambuf::init(char* mem, size_t size) {
	_mem = mem;
	_size = size;
	setg(_mem, _mem, _mem + _size);
}

std::streambuf::int_type sc::io::memstreambuf::underflow() {
	if (gptr() == _mem + _size) return EOF;
	return *gptr();
}

std::streambuf::int_type sc::io::memstreambuf::uflow() {
	if (gptr() == _mem + _size) return EOF;
	char c = *gptr();
	setg(_mem, gptr() + 1, _mem + _size);
	return c;
}

std::streambuf::int_type sc::io::memstreambuf::overflow(std::streambuf::int_type c) {
	// TODO
	// if (c != EOF) {
	// 	int res;
	// 	if ((res = write(_fd, &c, 1)) != 1) {
	// 		return EOF;
	// 	}
	// 	else {
	// 		std::cerr << "write result = " << res << ", errno = " << errno << '\n';
	// 	}
	// }
	// _ch = c;
	// setp(&_ch, &_ch + 1);
	return c;
}

std::streamsize sc::io::memstreambuf::xsputn(const char* /*s*/, std::streamsize num) {
	// TODO
// 	int res = write(_fd, s, num);
// 	if (res != num) {
// 		std::cerr << "write result = " << res << ", errno = " << errno << '\n';
// 		return EOF;
// 	}
	// return res;
	return num;
}

std::streambuf::pos_type sc::io::memstreambuf::seekoff(std::streambuf::off_type off,
													   std::ios_base::seekdir dir,
													   std::ios_base::openmode which) {
	if (which & std::ios::in) { // reading
		switch (dir) {
			case std::ios::beg:
				if (off < 0 || static_cast<size_t>(off) > _size) return EOF;
				setg(_mem, _mem + off, _mem + _size);
				break;
			case std::ios::cur: {
				int64_t pos = gptr() - _mem + off;
				if (pos < 0 || static_cast<size_t>(pos) > _size) return EOF;
				setg(_mem, gptr() + off, _mem + _size);
				break;
			}
			case std::ios::end:
				if (off > 0 || static_cast<size_t>(off) < -_size) return EOF;
				setg(_mem, _mem + _size + off - 1, _mem + _size);
				break;
		}
		return gptr() - _mem;
	} else if (which & std::ios::out) { // writing
		switch (dir) { // TODO
			case std::ios::beg:
				break;
			case std::ios::cur:
				break;
			case std::ios::end:
				break;
		}
		return pptr() - _mem;
	}
	return EOF;
}

std::streambuf::pos_type sc::io::memstreambuf::seekpos(std::streambuf::pos_type pos,
													   std::ios_base::openmode which) {
	if (which & std::ios::in) {
		if (pos < 0 || static_cast<size_t>(pos) > _size) return EOF;
		setg(_mem, _mem + pos, _mem + _size);
		return pos;
	} else if (which & std::ios::out) {
		// TODO
		return pos;
	}
	return EOF;
}

std::streambuf::int_type sc::io::memstreambuf::pbackfail(std::streambuf::int_type /*c*/) {
	return EOF;
}
